'''
    nicely tests.
'''
from util import check


check('test',
'''test
''')


check('test1\ntest2',
"""'''
    test1
    test2
'''
""")


check('test1\ntest2\ntest3\ntest4',
"""'''
    test1
    test2
    test3
    ... 4
'''
""",
max_lines = 2,
)


check('test 123456789',
'''test 1234 ...
''',
max_str = 13,
)


check('test1\ntest23456789\ntest3\ntest4\ntest5',
"""'''
    test1
    test23456 ...
    test3
    ... 5
'''
""",
max_lines = 2, max_str = 13,
)


check([3, 2, 1],
'''[
    1,
    2,
    3,
]
''',
sort = True,
)


check([3, 2, 1],
'''[
    3,
    2,
    1,
]
''',
sort = False,
)


check([3, 'a', 1],
'''[
    3,
    a,
    1,
]
''')


check([3, 'a\nb', 1],
"""[
    3,
    '''
        a
        b
    ''',
    1,
]
""")


check([1,[2],3],
'''[
    1,
    [
        2,
    ],
    3,
]
''')


check([1,2,[4,3],5,6],
'''[
    1,
    2,
    [
        3,
        4,
    ],
    5,
    6,
]
''',
sort = True,
)


check([1,2,[4,3],5,6],
'''[
    1,
    2,
    [
        4,
        3,
    ],
    5,
    6,
]
''',
sort = False,
)


check([1,2,[4,3],5,6],
'''[
    1,
    2,
    [
        3,
        4,
    ],
    ... 5
]
''',
max_seq = 2, sort = True,
)


check(dict(a=1, bb=2, ccc=3, d=4, e=5),
'''{
    a   = 1,
    bb  = 2,
    ccc = 3,
    d   = 4,
    e   = 5,
}
''',
sort = True,
)


check(dict(a=1, bb=2, ccc=3, d=4, e=5),
'''{
    a   = 1,
    bb  = 2,
    ccc = 3,
    ... 5
}
''',
max_map = 2, sort = True,
)


check(dict(a=1, bb=2, ccc=3, dddd=4, eeeee=5),
'''{
    a   = 1,
    bb  = 2,
    ccc = 3,
    ... 5
}
''',
max_map = 2, sort = True,
)


check(dict(a=1, bb=dict(d=9, eeeee=8), ccc=3),
'''{
    a   = 1,
    bb  = {
        d     = 9,
        eeeee = 8,
    },
    ccc = 3,
}
''',
sort = True,
)


check(dict(a=1, bb=[9, 8], ccc=3),
'''{
    a   = 1,
    bb  = [
        8,
        9,
    ],
    ccc = 3,
}
''',
sort = True,
)


check(dict(a=1, bb=[9, 8], ccc='''foo\nbar'''),
"""{
    a   = 1,
    bb  = [
        8,
        9,
    ],
    ccc = '''
        foo
        bar
    ''',
}
""",
sort = True
)


check([1, 2, dict(a=3, b=4), 5, 6],
'''[
    1,
    2,
    {
        a = 3,
        b = 4,
    },
    5,
    6,
]
''',
sort = True,
)


check([1, 2, dict(f=(999, 888, 777), a=3, bbb=4, c=dict(dddd=5, e=6)), 7, 8],
'''[
    1,
    2,
    {
        a   = 3,
        bbb = 4,
        c   = {
            dddd = 5,
            e    = 6,
        },
        f   = (
            777,
            888,
            999,
        ),
    },
    7,
    8,
]
''',
sort = True,
)


check([1, 2, dict(f=(999, 888, 777), a=3, bbb=4, c=dict(dddd=5, e=6)), 7, 8],
'''[
    1,
    2,
    {
        a   = 3,
        bbb = 4,
        ... 4
    },
    ... 5
]
''',
max_seq = 2, max_map = 1, sort = True,
)


from collections import namedtuple

nt = namedtuple('test', 'foo_bar baz')
check(nt(2, 3),
'''(
    baz     = 3,
    foo_bar = 2,
)
''',
sort = True,
)


nt2 = namedtuple('test2', 'foo bar baz')
check(nt2(2, 3, 5),
'''(
    bar = 3,
    baz = 5,
    foo = 2,
)
''',
sort = True,
)


check(nt2(2, 3, 5),
'''(
    bar = 3,
    foo = 2,
)
''',
sort = True, ignore='baz',
)


check(set((1, 5, frozenset((3, 2, 1)), 2, 6)),
'''{
    1,
    2,
    5,
    6,
    {
        1,
        2,
        3,
    },
}
''',
sort = True,
)


check(frozenset((1, 5, 2, 6)),
'''{
    1,
    2,
    5,
    6,
}
''',
sort = True,
)


check(frozenset('1 2 3 4 5 6 7 9'.split()),
'''{
    1,
    2,
    3,
    4,
    ... 8
}
''',
max_seq = 3, sort = True,
)


check(dict(abc=3, abd=6, cab=9, cba=8, cbc=7),
'''{
    cab = 9,
    cba = 8,
    cbc = 7,
}
''',
sort = True, ignore = r'ab.*',
)


check(dict(abc=3, bad=6, cab=9, cba=8, cbc=7),
'''{
    cab = 9,
}
''',
sort = True, ignore = (r'.*bc', r'.*ba'),
)


check([1,2,3,[4,5,6]],
'''[
|-->1,
|-->2,
|-->3,
|-->[
|-->|-->4,
|-->|-->5,
|-->|-->6,
|-->],
]
''',
indent = '|-->',
)


class Test_0:
    '''Printer Test Class'''
    def __init__(self):
        self.data = '''
            Foo
            Bar
            Ham
            Baz
        '''.split()
    def __print__(self, printer):
        with printer.brackets('<<<', '>>>'):
            printer.print(self.data)


check(Test_0(),
"""<<<
    [
        Bar,
        Baz,
        Foo,
        Ham,
    ]
>>>
""",
sort = True,
)


class Test_1:
    def __init__(self):
        self.data = '''
            Foo
            Bar
            Ham egg spam
            Baz
        '''
    def __print__(self, printer):
        with printer.brackets('<<<', '>>>'):
            printer.print(self.data)


check(Test_1(),
"""<<<
    '''
        Foo
        Bar
        Ham egg spam
        Baz
    '''
>>>
""")


check([123, Test_1(), 456],
"""[
    123,
    <<<
        '''
            Foo
            Bar
            Ham egg spam
            Baz
        '''
    >>>,
    456,
]
""")


class Test_1b:
    def __init__(self):
        self.data = ('''
            Foo
            Bar
            Ham egg spam
            Baz
        ''',
        333,
        )
    def __print__(self, printer):
        with printer.brackets('{{', '}}'):
            printer.print(self.data)


check(Test_1b(),
"""{{
    (
        '''
            Foo
            Bar
            Ham egg spam
            Baz
        ''',
        333,
    )
}}
""")


check(Test_1b(),
"""Test_1b!{{
    tuple!(
        str!'''
            Foo
            Bar
            Ham egg spam
            Baz
        ''',
        int!333,
    )
}}
""",
types = True)


class Test_2:
    def __init__(self):
        self.data = [
            8,
            7,
            'foo bar',
            nt(33, 42),
            set((3.3, 42)),
            3,
            [7, 8, {5:6,8:9}, 9],
            [45, 78, 12, 33],
            Test_1(),
            dict(a=3, b=7, c=Test_1(), d=[3,6,Test_1(),9], e=dict(ebbbbbb=6, ec=4, eaa=3.3)),
        ]
    def __print__(self, printer):
        printer.print(self.data)


check(Test_2(),
"""[
    8,
    7,
    foo bar,
    (
        baz     = 42,
        foo_bar = 33,
    ),
    {
        3.3,
        42,
    },
    3,
    [
        7,
        8,
        {
            5 = 6,
            8 = 9,
        },
        9,
    ],
    [
        12,
        33,
        45,
        78,
    ],
    <<<
        '''
            Foo
            Bar
            Ham egg spam
            Baz
        '''
    >>>,
    {
        a = 3,
        b = 7,
        c = <<<
            '''
                Foo
                Bar
                Ham egg spam
                Baz
            '''
        >>>,
        d = [
            3,
            6,
            <<<
                '''
                    Foo
                    Bar
                    Ham egg spam
                    Baz
                '''
            >>>,
            9,
        ],
        e = {
            eaa     = 3.3,
            ebbbbbb = 6,
            ec      = 4,
        },
    },
]
""",
sort = True,
)


check(Test_2(),
"""list![
    int!8,
    int!7,
    str!foo bar,
    test!(
        str!baz     = int!42,
        str!foo_bar = int!33,
    ),
    set!{
        float!3.3,
        int!42,
    },
    int!3,
    list![
        int!7,
        int!8,
        dict!{
            int!5 = int!6,
            int!8 = int!9,
        },
        int!9,
    ],
    list![
        int!12,
        int!33,
        int!45,
        int!78,
    ],
    Test_1!<<<
        str!'''
            Foo
            Bar
            Ham egg spam
            Baz
        '''
    >>>,
    dict!{
        str!a = int!3,
        str!b = int!7,
        str!c = Test_1!<<<
            str!'''
                Foo
                Bar
                Ham egg spam
                Baz
            '''
        >>>,
        str!d = list![
            int!3,
            int!6,
            Test_1!<<<
                str!'''
                    Foo
                    Bar
                    Ham egg spam
                    Baz
                '''
            >>>,
            int!9,
        ],
        str!e = dict!{
            str!eaa     = float!3.3,
            str!ebbbbbb = int!6,
            str!ec      = int!4,
        },
    },
]
""",
types = True, sort = True,
)


check('',
'''""
''')


check([''],
'''[
    "",
]
''')


check(['', 'a'],
'''[
    "",
    a,
]
''')


check([],
'''[
]
''')


check((),
'''(
)
''')


check({},
'''{
}
''')


check(dict(test='name'),
'''{
    test = name,
}
''',
name = 'test.name',
)


check(Test_0,
'''class:Test_0<
    __doc__           = Printer Test Class,
    __init__          = <function Test_0.__init__  ...,
    __init_subclass__ = <built-in method __init_su ...,
    __module__        = __main__,
    __print__         = <function Test_0.__print__ ...,
>
''',
max_str = 30, sort = True,
)


def check_math():
    import math
    check(math,
"""module:math<
    __doc__     = '''
        This module is always available.  It provides access to the
        mathematical functions defined by the C standard.
    ''',
    __name__    = math,
    __package__ = "",
    __spec__    = ModuleSpec(name='math', loader=<class '_frozen_importlib.BuiltinImporter'>,  ...,
    acos        = <built-in function acos>,
    acosh       = <built-in function acosh>,
    asin        = <built-in function asin>,
    asinh       = <built-in function asinh>,
    atan        = <built-in function atan>,
    atan2       = <built-in function atan2>,
    atanh       = <built-in function atanh>,
    ceil        = <built-in function ceil>,
    copysign    = <built-in function copysign>,
    cos         = <built-in function cos>,
    cosh        = <built-in function cosh>,
    degrees     = <built-in function degrees>,
    e           = 2.718281828459045,
    ... 53
>
""",
        max_map = 16, max_str = 80, sort = True,
    )


    check(math,
"""module:math<
    atan        = <built-in function atan>,
    atan2       = <built-in function atan2>,
    atanh       = <built-in function atanh>,
    ceil        = <built-in function ceil>,
    copysign    = <built-in function copysign>,
    degrees     = <built-in function degrees>,
    e           = 2.718281828459045,
    erf         = <built-in function erf>,
    erfc        = <built-in function erfc>,
    exp         = <built-in function exp>,
    expm1       = <built-in function expm1>,
    fabs        = <built-in function fabs>,
    factorial   = <built-in function factorial>,
    floor       = <built-in function floor>,
    fmod        = <built-in function fmod>,
    frexp       = <built-in function frexp>,
    fsum        = <built-in function fsum>,
    ... 53
>
""",
        max_map = 26, max_str = 80, sort = True, ignore = (r'__.*__', r'.*(sin|cos).*',),
    )


print(' ok')



''' nicely example

    Example uses of the nicely.Printer class.
'''

import collections

from nicely import Printer

'''
    The simplest way to use Printer.print is
'''
p = Printer()
p.print(collections)

'''
    or
'''
pp = Printer().print
pp(collections)

'''
    of course, we can change the config
'''
p = Printer(max_lines=33, max_map=33)
p.print(collections)

'''
    or the destination
'''
with open('collections', 'w') as out:
    pp = Printer(max_map=33, max_lines=33, target=out).print
    pp(collections.Counter)
    pp(collections.ChainMap)



import codecs
from io import StringIO

from .printer import Printer


def dump(what, **kw):
    ''' Dump a data structure to stdout and to a file.

    The output to stdout uses the default parameters.
    The parameters for output to the file may be specified as key-args.
    The name and the ext of the output file may be specified as well.
    The name defaults to "nicely" and the ext to "dump".
    '''
    if kw.pop('default', True):
        Printer().print(what)

    arg = dict(
        sort      = False,
        types     = False,
        max_seq   = 99,
        max_map   = 99,
        max_str   = 99,
        max_lines = 33,
        max_depth =  9,
    )
    name = kw.pop('name', 'nicely')
    name, ext = name.split('.') if '.' in name else (name, 'dump')
    arg.update(kw)
    with codecs.open(f'{name}.{ext}', 'w', encoding='utf-8') as out:
        Printer(target=out, **arg).print(what)


def format(obj, **kw):
    '''Format a data structure.

    Return a string representation of the object passed as arg.
    Formatting parameters may be specified as key-args.
    '''
    arg = dict(
        sort      = False,
        types     = False,
        max_seq   = 99,
        max_map   = 99,
        max_str   = 99,
        max_lines = 33,
        max_depth =  9,
    )
    arg.update(kw)
    out = StringIO()
    Printer(target=out, **arg).print(obj)
    string = out.getvalue()
    out.close()
    return string



''' nicely example

    This module shows how to implement the '__print__' method to make objects "printable".
'''

from nicely import dump

'''
    We need to define a '__print__' method for each class that we want to 'dump'.
'''

class Element:
    tag = ''
    def __init__(self, text):
        self.text = text
    @property
    def opening(self):
        return '<%s>'%self.tag
    @property
    def closing(self):
        return '</%s>'%self.tag
    def __print__(self, printer):
        with printer.brackets(self.opening, self.closing):
            printer.print(self.text)

class Section(Element):
    def __init__(self, *elements):
        self.elements = elements
    def __print__(self, printer):
        with printer.brackets(self.opening, self.closing):
            for e in self.elements:
                printer.print(e)

'''
    Now we can instanciate our classes, and 'dump' them.
'''

class Title(Element): tag = 'title'

class P(Element): tag = 'p'

class Head(Section): tag = 'head'

class Body(Section): tag = 'body'

class Html(Section): tag = 'html'

html = Html(
    Head(
        Title('Title'),
    ),
    Body(
        P('First para'),
        P('Second one'),
    ),
)

dump(html)



''' nicely example

    This module shows how to implement the __print__ function to make objects "printable".
'''

from nicely.formatters import Formatter
from nicely.printer import Printer
from nicely import dump

'''
    Let's say that we want to make the following classes "printable".
'''

class Element:
    tag = ''
    def __init__(self, text):
        self.text = text
    @property
    def opening(self):
        return '<%s>'%self.tag
    @property
    def closing(self):
        return '</%s>'%self.tag

class Section(Element):
    def __init__(self, *elements):
        self.elements = elements

'''
    For the Printer to be able to handle our classes
    we need to define a formatter for each one of them.
'''
class Element_formatter(Formatter):
    def validate(self):
        return isinstance(self.obj, Element)
    def __print__(self, printer):
        with printer.brackets(self.obj.opening, self.obj.closing):
            printer.print(self.obj.text)

class Section_formatter(Formatter):
    def validate(self):
        return isinstance(self.obj, Section)
    def __print__(self, printer):
        with printer.brackets(self.obj.opening, self.obj.closing):
            for e in self.obj.elements:
                printer.print(e)

'''
    We add our custom formatters to the Printer's ones.
    Note that as we use isinstance to 'validate',
    the order in which the formatters are appened is important.
'''
Printer.formatters.append(Section_formatter)
Printer.formatters.append(Element_formatter)

'''
    Now we can instanciate our custon classes,
    they will be handled by the Printer as expected.
'''

class Title(Element): tag = 'title'

class P(Element): tag = 'p'

class Head(Section): tag = 'head'

class Body(Section): tag = 'body'

class Html(Section): tag = 'html'

html = Html(
    Head(
        Title('Title'),
    ),
    Body(
        P('First para'),
        P('Second one'),
    ),
)

dump(html)



'''
    Data type Formaters.

    All Formaters must be derived from Formater.
'''

from types import ModuleType


#pylint: disable=invalid-name


class Formatter:
    ''' Base class for all Formaters. '''

    def __init__(self, obj):
        # The object to be formated.
        self.obj = obj

    def validate(self):
        '''
            Must be implemented by subclasses
            Should retun True if self.obj can be handled by this class.
        '''
        raise NotImplementedError

    def __print__(self, printer):
        '''
            Must be implemented by subclasses
            Call printer's methods appropriatly to get the desired result.
        '''
        raise NotImplementedError


'''
    Default Formaters for built-in container types and namespaces.
'''


class Tuple_formatter(Formatter):

    def validate(self):
        return isinstance(self.obj, tuple)

    def __print__(self, printer):
        with printer.brackets('(', ')'):
            printer.print_seq(self.obj)


class List_formatter(Formatter):

    def validate(self):
        return isinstance(self.obj, list)

    def __print__(self, printer):
        with printer.brackets('[', ']'):
            printer.print_seq(self.obj)


class Iter_formatter(List_formatter):

    def validate(self):
        return hasattr(self.obj, '__iter__') and not isinstance(self.obj, str)


class Dict_formatter(Formatter):

    def validate(self):
        return isinstance(self.obj, dict)

    def __print__(self, printer):
        with printer.brackets('{', '}'):
            printer.print_map(self.obj.items())


class Namedtuple_formatter(Formatter):

    def validate(self):
        return isinstance(self.obj, tuple) and getattr(self.obj, '_fields', False)

    def __print__(self, printer):
        with printer.brackets('(', ')'):
            printer.print_map(list(zip(self.obj._fields, self.obj)))


class Set_formatter(Formatter):

    def validate(self):
        return isinstance(self.obj, set)

    def __print__(self, printer):
        with printer.brackets('{', '}'):
            printer.print_seq(self.obj)


class Frozenset_formatter(Formatter):

    def validate(self):
        return isinstance(self.obj, frozenset)

    def __print__(self, printer):
        with printer.brackets('{', '}'):
            printer.print_seq(self.obj)


class Module_formatter(Formatter):

    def validate(self):
        if isinstance(self.obj, ModuleType):
            self.name = self.obj.__name__
            self.op_bk = 'module:%s<' % self.name
            self.cl_bk = '>'
            return True
        return False

    def __print__(self, printer):
        exclude = '__builtins__',
        items = {}
        for name in dir(self.obj):
            if name in exclude:
                continue
            else:
                val = getattr(self.obj, name)
                if isinstance(val, ModuleType):
                    if val.__name__.startswith(self.name):
                        items[name] = val
                elif isinstance(val, type):
                    if val.__module__.startswith(self.name):
                        items[name] = val
                else:
                    items[name] = val
        with printer.brackets(self.op_bk, self.cl_bk):
            printer.print_map(items.items())


class Class_formatter(Formatter):

    def validate(self):
        exclude = 'object', 'type'
        if isinstance(self.obj, type):
            name = self.obj.__name__
            if name in exclude:
                return False
            self.op_bk = 'class:%s<' % name
            self.cl_bk = '>'
            return True
        return False

    def __print__(self, printer):
        exclude = '__abstractmethods__', '__dict__', '__weakref__', '__new__', '__subclasshook__'
        nontypes = (
            'wrapper_descriptor', 'method_descriptor', 'member_descriptor',
            'getset_descriptor', 'BuiltinFunctionType', 'BuiltinMethodType',
        )
        items = {}
        for name in dir(self.obj):
            if name in exclude:
                continue
            else:
                val = getattr(self.obj, name)
                if type(val).__name__ in nontypes:
                    continue
                elif isinstance(val, type):
                    if val.__module__.startswith(self.obj.__module__):
                        items[name] = val
                else:
                    items[name] = val
        with printer.brackets(self.op_bk, self.cl_bk):
            printer.print_map(items.items())



''' nicely example

    Example uses ot the nicely.dump function.
'''

import collections

from nicely import dump

'''
    The simplest way to use dump is
'''
dump(collections)
'''
    which uses the default parameters and so dumps the argument on stdout and to the 'nicely.dump' file.
'''

'''
    In addition to all the arguments that the nicely.Printer class would accept,
    we can pass to the dump function the name of the destination file.
'''
dump(collections.Counter, name='counter')       # we can use the default ext (dump)
dump(collections.ChainMap, name='chainmap.log') # or specifiy the full name

'''
    if the output on the screen is not wanted, we can set the 'default' parameter to False.
'''
dump(collections, default=False)
'''
    if the name of the destination file is not specified, if will be 'nicely.dump'.
'''



'''
    Nice Printer
'''

from contextlib import contextmanager
from operator import itemgetter
from sys import stdout
import re

from .utils import State, Stack
#pylint: disable=wildcard-import,unused-wildcard-import
from .formatters import *


#pylint: disable=unused-variable


#pylint: disable=too-many-instance-attributes
class Printer:
    ''' Fancy Printer

    Dump an arbritrary data structure to a file-like destination.
    '''

    # One formater is needed for each supported type.
    # Strings don't need a Formater as there're handled directly by the Printer.
    # User Formaters may be added or inserted.
    # Watch the order, the first matching Formater will be used.
    formatters = [
        Namedtuple_formatter,
        Tuple_formatter,
        List_formatter,
        Dict_formatter,
        Set_formatter,
        Frozenset_formatter,
        Iter_formatter,
        Module_formatter,
        Class_formatter,
    ]
    # All formatters must subclass Formatter
    assert all(issubclass(f, Formatter) for f in formatters)

    #pylint: disable=too-many-arguments,line-too-long
    def __init__(self, sort=True, types=False, max_seq=9, max_map=9, max_str=77, max_lines=9, max_depth=9, ignore=None, indent=None, target=stdout):
        self.sort = sort                  # if True: sequences will be sorted (if possible)
        self.max_seq = max_seq            # max items in sequence
        self.max_map = max_map            # max items in mapping
        self.max_str = max(max_str-4, 4)  # max string length (at least 4)
        self.max_lines = max_lines        # max lines in multi-lines string
        self.max_depth = max_depth        # max structure depth
        self.ignore = ignore or []        # keys to ignore in mappings
        self.indent = indent or '    '    # indentation string
        self.target = target.write        # output destination
        self.types = State(types)         # if True: data types will be shown
        ignore = [self.ignore] if isinstance(self.ignore, str) else self.ignore
        self._ignore = [re.compile(r).match for r in ignore]
        self._indent = State(True)
        self._inline = State(False)
        self._width = State(1)
        self._stack = Stack()
        self._level = 0

    def wrap(self, obj):
        # Wrap obj with the appropriate Formatter.
        #pylint: disable=invalid-name,redefined-outer-name
        for Formatter in self.formatters:
            formatter = Formatter(obj)
            if formatter.validate():
                return formatter
        return obj

    @contextmanager
    def brackets(self, op_bk, cl_bk):
        # Handle opening and closing tags.
        if self.types:
            parent = self._stack.last
            typ = type(parent.obj) if isinstance(parent, Formatter) else type(parent)
            typ = typ.__name__+'!'
        else:
            typ = ''
        with self.types(False), self._inline(False):
            self.print_obj(typ+op_bk)
        self._level += 1
        with self._indent(True), self._inline(False):
            yield
        self._level -= 1
        with self.types(False), self._indent(True):
            self.print_obj(cl_bk)

    def print_obj(self, arg=''):
        # Main printing method.
        arg = self.wrap(arg)
        if hasattr(arg, '__print__'):  # That's a Formatter.
            with self._stack(arg):
                arg.__print__(self)
        else:
            ind = self.indent*self._level if self._indent else ''
            typ = '{}!'.format(type(arg).__name__) if self.types else ''
            pad = self._width.value
            ret = '' if self._inline else '\n'
            arg = str(arg)
            if len(arg) > self.max_str:
                arg = '%s ...' % arg[:self.max_str]
            self.target('{ind}{typ}{arg:{pad}}{ret}'.format(**vars()))

    def print_str_obj(self, arg=''):
        # Send arg to the appropriate handler.
        #pylint: disable=expression-not-assigned
        self.print_str(arg) if isinstance(arg, str) else self.print_obj(arg)

    @contextmanager
    def trailing_comma(self):
        # add a comma a the end of a line in a sequence.
        with self._inline(True):
            yield
        with self._indent(False), self._inline(False), self.types(False):
            self.print_obj(',')

    def check_depth(self):
        # Limit output to max_depth.
        if len(self._stack) > self.max_depth:
            with self.types(False), self._inline(False):
                self.print_obj('...')
            return False
        return True

    def sort_seq(self, seq):
        # sort seq if needed and possible.
        seq = list(seq)
        if self.sort:
            try:
                seq.sort()
            except TypeError: pass
        return seq

    def limit(self, seq, lim):
        # check the limits.
        for cpt, item in enumerate(seq):
            if cpt > lim:
                with self.types(False), self._inline(False):
                    self.print_obj('... %s' % len(seq))
                break
            yield item

    def print_seq(self, seq):
        # Handle sequences.
        if not self.check_depth(): return
        seq = self.sort_seq(seq)
        for item in self.limit(seq, self.max_seq):
            with self.trailing_comma():
                self.print_str_obj(item)

    def print_map(self, items):
        # Handle mappings.
        if not items: return
        if not self.check_depth(): return
        items = self.sort_seq(items)
        #pylint: disable=bad-builtin
        width = max(map(len, map(str, map(itemgetter(0), items[:self.max_map+1]))))
        for key, val in self.limit(items, self.max_map):
            if any(match(key) for match in self._ignore):
                continue
            with self.trailing_comma():
                with self._width(width):
                    self.print_obj(key)
                with self._indent(False):
                    with self.types(False):
                        self.print_obj(' = ')
                    self.print_str_obj(val)

    def print_str(self, arg):
        # Handle strings.
        lines = [lin for lin in (line.strip() for line in arg.split('\n')) if lin]
        if len(lines) > 1:
            with self._stack(arg), self.brackets("'''", "'''"):
                for line in self.limit(lines, self.max_lines):
                    with self._inline(False), self.types(False):
                        self.print_obj(line)
        else:
            self.print_obj(lines[0] if lines else '""')

    def print(self, arg):
        # Entry point.
        self._indent.value = True
        self.print_str_obj(arg)



'''
    Utility classes.
'''

from contextlib import contextmanager


class State:
    ''' Save state on enter and restore it on exit '''

    def __init__(self, value):
        self.value = value

    @contextmanager
    def _cm(self, val):
        bak = self.value
        self.value = val
        yield
        self.value = bak

    def __call__(self, val):
        return self._cm(val)

    def __bool__(self):
        return bool(self.value)


class Stack:
    ''' Push a value on enter and pop it on exit'''

    def __init__(self):
        self._stack = []

    @contextmanager
    def _cm(self, val):
        self._stack.append(val)
        yield
        self._stack.pop()

    def __call__(self, val):
        return self._cm(val)

    def __len__(self):
        return len(self._stack)

    @property
    def last(self):
        return self._stack[-1]
